"""Returns tidy3d simulation from gdsfactory Component."""
import warnings
from typing import Dict, Optional

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pydantic
import tidy3d as td

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.extension import move_polar_rad_copy
from gdsfactory.config import logger
from gdsfactory.pdk import get_layer_stack
from gdsfactory.routing.sort_ports import sort_ports_x, sort_ports_y
from gdsfactory.simulation.gtidy3d.materials import (
    MATERIAL_NAME_TO_TIDY3D_INDEX,
    MATERIAL_NAME_TO_TIDY3D_NAME,
    get_index,
    get_medium,
)
from gdsfactory.tech import LayerStack
from gdsfactory.types import ComponentSpec, Float2


@pydantic.validate_arguments
def get_simulation(
    component: ComponentSpec,
    port_extension: Optional[float] = 4.0,
    layer_stack: Optional[LayerStack] = None,
    thickness_pml: float = 1.0,
    xmargin: float = 0,
    ymargin: float = 0,
    xmargin_left: float = 0,
    xmargin_right: float = 0,
    ymargin_top: float = 0,
    ymargin_bot: float = 0,
    zmargin: float = 1.0,
    clad_material: str = "sio2",
    port_source_name: str = "o1",
    port_margin: float = 0.5,
    port_source_offset: float = 0.1,
    distance_source_to_monitors: float = 0.2,
    wavelength_start: float = 1.50,
    wavelength_stop: float = 1.60,
    wavelength_points: int = 50,
    plot_modes: bool = False,
    num_modes: int = 2,
    run_time_ps: float = 10.0,
    dispersive: bool = False,
    material_name_to_tidy3d_index: Dict[str, float] = MATERIAL_NAME_TO_TIDY3D_INDEX,
    material_name_to_tidy3d_name: Dict[str, str] = MATERIAL_NAME_TO_TIDY3D_NAME,
    is_3d: bool = True,
    with_all_monitors: bool = False,
    boundary_spec: Optional[td.BoundarySpec] = None,
    grid_spec: Optional[td.GridSpec] = None,
    sidewall_angle_deg: float = 0,
    dilation: float = 0.0,
    **kwargs,
) -> td.Simulation:
    r"""Returns tidy3d Simulation object from a gdsfactory Component.

    based on GDS example
    https://simulation.cloud/docs/html/examples/ParameterScan.html

    .. code::

         top view
              ________________________________
             |                               |
             | xmargin_left                  | port_extension
             |<------>          port_margin ||<-->
          ___|___________          _________||___
             |           \        /          |
             |            \      /           |
             |             ======            |
             |            /      \           |
          ___|___________/        \__________|___
             |   |                 <-------->|
             |   |ymargin_bot   xmargin_right|
             |   |                           |
             |___|___________________________|

        side view
              ________________________________
             |                     |         |
             |                     |         |
             |                   zmargin_top |
             |xmargin_left         |         |
             |<---> _____         _|___      |
             |     |     |       |     |     |
             |     |     |       |     |     |
             |     |_____|       |_____|     |
             |       |                       |
             |       |                       |
             |       |zmargin_bot            |
             |       |                       |
             |_______|_______________________|


    Args:
        component: gdsfactory Component.
        port_extension: extend ports beyond the PML.
        layer_stack: contains layer to thickness, zmin and material.
            Defaults to active pdk.layer_stack.
        thickness_pml: PML thickness (um).
        xmargin: left/right distance from component to PML.
        xmargin_left: left distance from component to PML.
        xmargin_right: right distance from component to PML.
        ymargin: left/right distance from component to PML.
        ymargin_top: top distance from component to PML.
        ymargin_bot: bottom distance from component to PML.
        zmargin: thickness for cladding above and below core.
        clad_material: material for cladding.
        port_source_name: input port name.
        port_margin: margin on each side of the port.
        distance_source_to_monitors: in (um) source goes before monitors.
        port_source_offset: mode solver workaround.
            positive moves source forward, negative moves source backward.
        wavelength_start: in (um).
        wavelength_stop: in (um).
        wavelength_points: number of wavelengths.
        plot_modes: plot source modes.
        num_modes: number of modes to plot.
        run_time_ps: make sure it's sufficient for the fields to decay.
            defaults to 10ps and counts on automatic shutoff to stop earlier if needed.
        dispersive: False uses constant refractive index materials.
            True adds wavelength depending materials.
            Dispersive materials require more computation.
        material_name_to_tidy3d_index: not dispersive materials have a constant index.
        material_name_to_tidy3d_name: dispersive materials have a wavelength
            dependent index. Maps layer_stack names with tidy3d material database names.
        is_3d: if False, does not consider Z dimension for faster simulations.
        with_all_monitors: True includes field monitors which increase results filesize.
        grid_spec: defaults to automatic td.GridSpec.auto(wavelength=wavelength)
            td.GridSpec.uniform(dl=20*nm)
            td.GridSpec(
                grid_x = td.UniformGrid(dl=0.04),
                grid_y = td.AutoGrid(min_steps_per_wvl=20),
                grid_z = td.AutoGrid(min_steps_per_wvl=20),
                wavelength=wavelength,
                override_structures=[refine_box]
            )
        boundary_spec: Specification of boundary conditions along each dimension.
            Defaults to td.BoundarySpec.all_sides(boundary=td.PML())

        dilation: float = 0.0
            Dilation of the polygon in the base by shifting each edge along its
            normal outwards direction by a distance;
            a negative value corresponds to erosion.
        sidewall_angle_deg : float = 0
            Angle of the sidewall.
            ``sidewall_angle=0`` (default) specifies vertical wall,
            while ``0<sidewall_angle_deg<90`` for the base to be larger than the top.

    keyword Args:
        symmetry: Define Symmetries.
            Tuple of integers defining reflection symmetry across a plane
            bisecting the simulation domain normal to the x-, y-, and z-axis
            at the simulation center of each axis, respectvely.
            Each element can be `0` (no symmetry), `1` (even, i.e. 'PMC' symmetry) or
            `-1` (odd, i.e. 'PEC' symmetry).
            Note that the vectorial nature of the fields must be taken into account
            to correctly determine the symmetry value.
        medium: Background medium of simulation, defaults to vacuum if not specified.
        shutoff: shutoff condition
            Ratio of the instantaneous integrated E-field intensity to the maximum value
            at which the simulation will automatically terminate time stepping.
            prevents extraneous run time of simulations with fully decayed fields.
            Set to ``0`` to disable this feature.
        subpixel: averaging. True uses permittivity subpixel averaging,
            results in much higher accuracy for a given grid size.
        courant: Courant stability factor, controls time step to spatial step ratio.
            Lower values lead to more stable simulations for dispersive materials
            but simulations take longer.
        version: String specifying the front end version number.


    .. code::

        import matplotlib.pyplot as plt
        import gdsfactory as gf
        import gdsfactory.simulation.tidy3d as gt

        c = gf.components.bend_circular()
        sim = gt.get_simulation(c)
        gt.plot_simulation(sim)

    """
    component = gf.get_component(component)
    assert isinstance(component, Component)

    layer_stack = layer_stack or get_layer_stack()
    boundary_spec = boundary_spec or td.BoundarySpec.all_sides(boundary=td.PML())

    wavelength = (wavelength_start + wavelength_stop) / 2
    grid_spec = grid_spec or td.GridSpec.auto(wavelength=wavelength)

    layer_to_thickness = layer_stack.get_layer_to_thickness()
    layer_to_material = layer_stack.get_layer_to_material()
    layer_to_zmin = layer_stack.get_layer_to_zmin()
    # layer_to_sidewall_angle = layer_stack.get_layer_to_sidewall_angle()

    if dispersive:
        material_name_to_tidy3d = material_name_to_tidy3d_name
    else:
        material_name_to_tidy3d = material_name_to_tidy3d_index

    assert isinstance(
        component, Component
    ), f"component needs to be a gf.Component, got Type {type(component)}"
    if port_source_name not in component.ports:
        warnings.warn(
            f"port_source_name={port_source_name!r} not in {list(component.ports.keys())}"
        )
        port_source = component.get_ports_list(port_type="optical")[0]
        port_source_name = port_source.name
        warnings.warn(f"Selecting port_source_name={port_source_name!r} instead.")

    component_padding = gf.add_padding_container(
        component,
        default=0,
        top=ymargin or ymargin_top,
        bottom=ymargin or ymargin_bot,
        left=xmargin or xmargin_left,
        right=xmargin or xmargin_right,
    )
    component_extended = (
        gf.components.extend_ports(
            component=component_padding, length=port_extension, centered=True
        )
        if port_extension
        else component_padding
    )

    component_extended = component_extended.flatten()
    component_extended.name = component.name
    gf.show(component_extended)

    component_ref = component_padding.ref()
    component_ref.x = 0
    component_ref.y = 0

    clad_material_name_or_index = material_name_to_tidy3d[clad_material]
    clad = td.Structure(
        geometry=td.Box(
            size=(td.inf, td.inf, td.inf),
            center=(0, 0, 0),
        ),
        medium=get_medium(name_or_index=clad_material_name_or_index),
    )
    structures = [clad]

    layers_thickness = [
        layer_to_thickness[layer]
        for layer in component.get_layers()
        if layer in layer_to_thickness
    ]

    if len(layer_to_thickness) < 1:
        raise ValueError(f"{component.get_layers()} not in {layer_to_thickness.keys()}")

    t_core = max(layers_thickness)
    cell_thickness = (
        thickness_pml + t_core + thickness_pml + 2 * zmargin if is_3d else 0
    )

    sim_size = [
        component_ref.xsize + 2 * thickness_pml,
        component_ref.ysize + 2 * thickness_pml,
        cell_thickness,
    ]

    for layer in component.layers:
        if layer in layer_to_thickness and layer in layer_to_material:
            thickness = layer_to_thickness[layer]
            zmin = layer_to_zmin[layer] if is_3d else 0
            zmax = zmin + thickness if is_3d else 0

            if layer_to_material[layer] in material_name_to_tidy3d:
                name_or_index = material_name_to_tidy3d[layer_to_material[layer]]
                medium = get_medium(name_or_index=name_or_index)
                index = get_index(name_or_index=name_or_index)
                logger.debug(
                    f"Add {layer}, {name_or_index!r}, index = {index:.3f}, "
                    f"thickness = {thickness}, zmin = {zmin}, zmax = {zmax}"
                )

                polygons = td.PolySlab.from_gds(
                    gds_cell=component_extended,
                    gds_layer=layer[0],
                    gds_dtype=layer[1],
                    axis=2,
                    slab_bounds=(zmin, zmax),
                    sidewall_angle=np.deg2rad(sidewall_angle_deg),
                    dilation=dilation,
                )

                for polygon in polygons:
                    geometry = td.Structure(geometry=polygon, medium=medium)
                    structures.append(geometry)
            elif layer not in layer_to_material:
                logger.debug(f"Layer {layer} not in {list(layer_to_material.keys())}")
            else:
                materials = list(material_name_to_tidy3d.keys())
                logger.debug(f"material {layer_to_material[layer]} not in {materials}")

    # Add source
    port = component_ref.ports[port_source_name]
    angle = port.orientation
    width = port.width + 2 * port_margin
    size_x = width * abs(np.sin(angle * np.pi / 180))
    size_y = width * abs(np.cos(angle * np.pi / 180))
    size_x = 0 if size_x < 0.001 else size_x
    size_y = 0 if size_y < 0.001 else size_y
    size_z = cell_thickness - 2 * zmargin if is_3d else td.inf

    source_size = [size_x, size_y, size_z]
    source_center = port.center.tolist() + [0]  # (x, y, z=0)

    xy_shifted = move_polar_rad_copy(
        np.array(port.center), angle=angle * np.pi / 180, length=port_source_offset
    )
    source_center_offset = xy_shifted.tolist() + [0]  # (x, y, z=0)

    wavelengths = np.linspace(wavelength_start, wavelength_stop, wavelength_points)
    freqs = td.constants.C_0 / wavelengths
    freq0 = td.constants.C_0 / np.mean(wavelengths)
    fwidth = freq0 / 10

    msource = td.ModeSource(
        size=source_size,
        center=source_center,
        source_time=td.GaussianPulse(freq0=freq0, fwidth=fwidth),
        direction="-" if int(angle) in {0, 90} else "+",
    )

    # Add port monitors
    monitors = {}
    ports = sort_ports_x(sort_ports_y(component_ref.get_ports_list()))
    for port in ports:
        port_name = port.name
        angle = port.orientation
        width = port.width + 2 * port_margin
        size_x = width * abs(np.sin(angle * np.pi / 180))
        size_y = width * abs(np.cos(angle * np.pi / 180))
        size_x = 0 if size_x < 0.001 else size_x
        size_y = 0 if size_y < 0.001 else size_y
        size = (size_x, size_y, size_z)

        # if monitor has a source move monitor inwards
        length = -distance_source_to_monitors if port_name == port_source_name else 0
        xy_shifted = move_polar_rad_copy(
            np.array(port.center), angle=angle * np.pi / 180, length=length
        )
        center = xy_shifted.tolist() + [0]  # (x, y, z=0)

        monitors[port_name] = td.ModeMonitor(
            center=center,
            size=size,
            freqs=freqs,
            mode_spec=td.ModeSpec(num_modes=1),
            name=port.name,
        )

    zcenter = (zmax + zmin) / 2 if is_3d else 0
    domain_monitor = td.FieldMonitor(
        center=[0, 0, zcenter],
        size=[sim_size[0], sim_size[1], 0] if is_3d else [td.inf, td.inf, 0],
        freqs=[freq0],
        name="field",
    )
    monitors = list(monitors.values())
    monitors += [domain_monitor] if with_all_monitors else []

    sim = td.Simulation(
        size=sim_size,
        structures=structures,
        sources=[msource],
        monitors=monitors,
        run_time=run_time_ps * 1e-12,
        boundary_spec=boundary_spec,
        **kwargs,
    )

    if plot_modes:
        mode_spec = td.ModeSpec(num_modes=num_modes)
        src_plane = td.Box(center=source_center_offset, size=source_size)
        ms = td.plugins.ModeSolver(
            simulation=sim, plane=src_plane, freqs=[freq0], mode_spec=mode_spec
        )
        modes = ms.solve()
        print("Effective index of computed modes: ", np.array(modes.n_eff))

        if is_3d:
            fig, axs = plt.subplots(num_modes, 2, figsize=(12, 12))
            for mode_ind in range(num_modes):
                ms.plot_field(
                    "Ey", "abs", f=freq0, mode_index=mode_ind, ax=axs[mode_ind, 0]
                )
                ms.plot_field(
                    "Ez", "abs", f=freq0, mode_index=mode_ind, ax=axs[mode_ind, 1]
                )
        else:
            fig, axs = plt.subplots(num_modes, 3, figsize=(12, 12))
            for mode_ind in range(num_modes):
                ax1 = axs[mode_ind, 0]
                ax2 = axs[mode_ind, 1]
                ax3 = axs[mode_ind, 2]

                abs(modes.fields.Ex.sel(mode_index=mode_ind).abs).plot(ax=ax1)
                abs(modes.fields.Ey.sel(mode_index=mode_ind).abs).plot(ax=ax2)
                abs(modes.fields.Ez.sel(mode_index=mode_ind).abs).plot(ax=ax3)

                ax1.set_title(f"|Ex|: mode_index={mode_ind}")
                ax2.set_title(f"|Ey|: mode_index={mode_ind}")
                ax3.set_title(f"|Ez|: mode_index={mode_ind}")

        plt.show()
    return sim


def plot_simulation_yz(
    sim: td.Simulation,
    z: float = 0.0,
    y: float = 0.0,
    wavelength: Optional[float] = 1.55,
    figsize: Float2 = (11, 4),
):
    """Returns Simulation visual representation. Returns two views for 3D component and one view for 2D.

    Args:
        sim: simulation object.
        z: (um).
        y: (um).
        wavelength: (um) for epsilon plot if None plot structures.
        figsize: figure size.

    """
    fig = plt.figure(figsize=figsize)
    if sim.size[2] > 0.1 and sim.size[1] > 0.1:
        gs = mpl.gridspec.GridSpec(1, 2, figure=fig, width_ratios=[1, 1.4])
        ax1 = fig.add_subplot(gs[0, 0])
        ax2 = fig.add_subplot(gs[0, 1])
        if wavelength:
            freq = td.constants.C_0 / wavelength
            sim.plot_eps(z=z, ax=ax1, freq=freq)
            sim.plot_eps(y=y, ax=ax2, freq=freq)
        else:
            sim.plot(z=z, ax=ax1)
            sim.plot(y=y, ax=ax2)
    elif sim.size[2] > 0.1:  # 2D grating sim_size_y = 0
        gs = mpl.gridspec.GridSpec(1, 1, figure=fig, width_ratios=[1])
        ax1 = fig.add_subplot(gs[0, 0])
        if wavelength:
            freq = td.constants.C_0 / wavelength
            sim.plot_eps(y=y, ax=ax1, freq=freq)
        else:
            sim.plot(y=y, ax=ax1)

    else:  # 2D planar component size_z = 0
        gs = mpl.gridspec.GridSpec(1, 1, figure=fig, width_ratios=[1])
        ax1 = fig.add_subplot(gs[0, 0])
        if wavelength:
            freq = td.constants.C_0 / wavelength
            sim.plot_eps(z=z, ax=ax1, freq=freq)
        else:
            sim.plot(z=z, ax=ax1)

    plt.show()
    return fig


def plot_simulation_xz(
    sim: td.Simulation,
    x: float = 0.0,
    z: float = 0.0,
    wavelength: Optional[float] = 1.55,
    figsize: Float2 = (11, 4),
):
    """Returns figure with two axis of the Simulation.

    Args:
        sim: simulation object.
        x: (um).
        z: (um).
        wavelength: (um) for epsilon plot if None plot structures.
        figsize: figure size.

    """
    fig = plt.figure(figsize=figsize)
    gs = mpl.gridspec.GridSpec(1, 2, figure=fig, width_ratios=[1, 1.4])
    ax1 = fig.add_subplot(gs[0, 0])
    ax2 = fig.add_subplot(gs[0, 1])
    if wavelength:
        freq = td.constants.C_0 / wavelength
        sim.plot_eps(z=z, ax=ax1, freq=freq)
        sim.plot_eps(x=x, ax=ax2, freq=freq)
    else:
        sim.plot(z=z, ax=ax1)
        sim.plot(x=x, ax=ax2)
    plt.show()
    return fig


plot_simulation = plot_simulation_yz


if __name__ == "__main__":
    c = gf.components.mmi1x2()
    # c = gf.components.bend_circular(radius=2)
    # c = gf.components.crossing()
    # c = gf.c.straight_rib()

    c = gf.c.straight(length=3)
    sim = get_simulation(c, plot_modes=True, is_3d=True, sidewall_angle_deg=30)

    # sim = get_simulation(c, dilation=-0.2, is_3d=False)

    # sim = get_simulation(c, is_3d=True)
    # plot_simulation(sim)

    # filepath = pathlib.Path(__file__).parent / "extra" / "wg2d.json"
    # filepath.write_text(sim.json())

    # sim.plotly(z=0)
    # plot_simulation_yz(sim, wavelength=1.55)
    # fig = plt.figure(figsize=(11, 4))
    # gs = mpl.gridspec.GridSpec(1, 2, figure=fig, width_ratios=[1, 1.4])
    # ax1 = fig.add_subplot(gs[0, 0])
    # ax2 = fig.add_subplot(gs[0, 1])
    # sim.plot(z=0.0, ax=ax1)
    # sim.plot(x=0.0, ax=ax2)
    # plt.show()



from typing import Dict, Optional, Tuple

import numpy as np
import pygmsh

import gdsfactory as gf
from gdsfactory.geometry.boolean import boolean
from gdsfactory.pdk import get_layer_stack
from gdsfactory.tech import LayerStack
from gdsfactory.types import ComponentOrReference, Layer


def get_xsection_bounds_inplane(
    component: ComponentOrReference,
    xsection_bounds: Tuple[Tuple[float, float], Tuple[float, float]],
    line_layer: Optional[Layer] = 99,
    line_width: Optional[float] = 0.01,
):
    """Given a component c and two coordinates (x1,y1), (x2,y2), computes the \
        bounding box(es) of each layer in the xsection coordinate system (u).

    Uses gdsfactory boolean and component.extract

    Args:
        component: Component or ComponentReference.
        xsection_bounds: ( (x1,y1), (x2,y2) ), with x1,y1 beginning point of cross-sectional line and x2,y2 the end.
        line_layer: (dummy) layer to put the extraction line on.
        line_width: (dummy) thickness of extraction line. Cannot be 0, should be small (near dbu) for accuracy.

    Returns: Dict containing layer(list pairs, with list a list of bounding box coordinates (u1,u2))
        in xsection line coordinates.
    """
    # Create line component for bool
    P = gf.Path(xsection_bounds)
    X = gf.CrossSection(width=line_width, layer=line_layer)
    line = gf.path.extrude(P, X)

    # Ref line vector
    ref_vector_origin = np.array(xsection_bounds[0])

    # For each layer in component, extract point where line intersects
    # Choose intersecting polygon edge with maximal (including sign) dot product to reference line
    bounds_dict = {}
    for layer in component.get_layers():
        bounds_layer = []
        c_layer = component.extract(layer)
        layer_boolean = boolean(c_layer, line, "and")
        polygons = layer_boolean.get_polygons()
        for polygon in polygons:
            # Min point is midpoint between two closest vertices to origin
            distances_origin = np.linalg.norm(polygon - ref_vector_origin, axis=1)
            distances_sort_index = np.argsort(distances_origin)
            min_point = (
                polygon[distances_sort_index[0]] + polygon[distances_sort_index[1]]
            ) / 2
            max_point = (
                polygon[distances_sort_index[2]] + polygon[distances_sort_index[3]]
            ) / 2

            umin = np.linalg.norm(min_point - ref_vector_origin)
            umax = np.linalg.norm(max_point - ref_vector_origin)

            bounds_layer.append((umin, umax))
        bounds_dict[layer] = bounds_layer

    return bounds_dict


def get_xsection_bounds(
    component: ComponentOrReference,
    xsection_bounds: Tuple[Tuple[float, float], Tuple[float, float]],
    layer_stack: Optional[LayerStack] = None,
    exclude_layers: Optional[Tuple[Layer, ...]] = None,
):
    """Given a component and layer stack, computes the bounding box(es) of each \
            layer in the xsection coordinate system (u,z).

    Args:
        component: Component or ComponentReference.
        xsection_bounds: ( (x1,y1), (x2,y2) ), with x1,y1 beginning point of cross-sectional line and x2,y2 the end
        line_layer: (dummy) layer to put the extraction line on.
        line_width: (dummy) thickness of extraction line. Cannot be 0, should be small (near dbu) for accuracy.

    Returns: Dict containing layer: list pairs, with list a list of bounding box coordinates (u1,u2) in xsection line coordinates
    """
    # Get in-plane cross-sections
    inplane_bounds_dict = get_xsection_bounds_inplane(component, xsection_bounds)

    # Generate full bounding boxes
    layer_stack = layer_stack or get_layer_stack()
    layer_to_thickness = layer_stack.get_layer_to_thickness()
    layer_to_zmin = layer_stack.get_layer_to_zmin()
    exclude_layers = exclude_layers or ()

    outplane_bounds_dict = {}

    for layer, inplane_bounds_list in inplane_bounds_dict.items():
        outplane_bounds_list = []
        for inplane_bounds in inplane_bounds_list:
            if (
                layer not in exclude_layers
                and layer in layer_to_thickness
                and layer in layer_to_zmin
            ):
                height = layer_to_thickness[layer]
                zmin = layer_to_zmin[layer]

                # Get bounding box
                umin = np.min(inplane_bounds)
                umax = np.max(inplane_bounds)
                zmax = zmin + height
                outplane_bounds_list.append(
                    {"umin": umin, "umax": umax, "zmin": zmin, "zmax": zmax}
                )

        outplane_bounds_dict[layer] = outplane_bounds_list

    return outplane_bounds_dict


def mesh2D(
    component: ComponentOrReference,
    xsection_bounds: Tuple[Tuple[float, float], Tuple[float, float]],
    base_resolution: float = 0.2,
    refine_resolution: Optional[Dict[Layer, float]] = None,
    padding: Tuple[float, float, float, float] = (2.0, 2.0, 2.0, 2.0),
    layer_stack: Optional[LayerStack] = None,
    exclude_layers: Optional[Tuple[Layer, ...]] = None,
):
    """Returns gmsh 2D geometry of component along cross-sectional line (x1,y1), (x2,y2).

    Args:
        component: Component or ComponentReference.
        xsection_bounds: ( (x1,y1), (x2,y2) ), with x1,y1 beginning point of cross-sectional line and x2,y2 the end.
        base_resolution: background mesh resolution (um).
        refine_resolution: feature mesh resolution (um); layer dependent via a dict (default to base_resolution).
        padding: amount (left, right, bottom, up) to enlarge simulation region beyond features (um).

    """
    layer_stack = layer_stack or get_layer_stack()
    layer_to_thickness = layer_stack.get_layer_to_thickness()
    layer_to_zmin = layer_stack.get_layer_to_zmin()
    exclude_layers = exclude_layers or ()

    geometry = pygmsh.geo.Geometry()

    model = geometry.__enter__()

    # Find extremal coordinates
    bounds_dict = get_xsection_bounds(component, xsection_bounds)
    polygons = [polygon for polygons in bounds_dict.values() for polygon in polygons]
    umin = min(polygon["umin"] for polygon in polygons) - padding[0]
    umax = max(polygon["umax"] for polygon in polygons) + padding[1]
    zmin = min(polygon["zmin"] for polygon in polygons) - padding[2]
    zmax = max(polygon["zmax"] for polygon in polygons) + padding[3]

    # Background oxide
    points = [
        model.add_point([umin, zmin], mesh_size=base_resolution),
        model.add_point([umax, zmin], mesh_size=base_resolution),
        model.add_point([umax, zmax], mesh_size=base_resolution),
        model.add_point([umin, zmax], mesh_size=base_resolution),
    ]
    channel_lines = [
        model.add_line(points[i], points[i + 1]) for i in range(-1, len(points) - 1)
    ]
    channel_loop = model.add_curve_loop(channel_lines)

    # Add layers
    blocks = []
    for layer in component.get_layers():
        if (
            layer not in exclude_layers
            and layer in layer_to_thickness
            and layer in layer_to_zmin
        ):
            if bounds_dict[layer] == []:
                continue
            blocks_layer = []
            for i, bounds in enumerate(bounds_dict[layer]):
                points = [
                    [bounds["umin"], bounds["zmin"]],
                    [bounds["umin"], bounds["zmax"]],
                    [bounds["umax"], bounds["zmax"]],
                    [bounds["umax"], bounds["zmin"]],
                ]
                polygon = model.add_polygon(
                    points,
                    mesh_size=refine_resolution[layer]
                    if refine_resolution
                    else base_resolution,
                )
                model.add_physical(polygon, f"{layer}_{i}")
                blocks.append(polygon)
                blocks_layer.append(polygon)
            model.add_physical(blocks_layer, f"{layer}")
    plane_surface = model.add_plane_surface(channel_loop, holes=blocks)

    model.add_physical(plane_surface, "oxide")
    model.add_physical([channel_lines[0]], "left")
    model.add_physical([channel_lines[1]], "bottom")
    model.add_physical([channel_lines[2]], "right")
    model.add_physical([channel_lines[3]], "top")

    geometry.generate_mesh(dim=2, verbose=True)

    return geometry


if __name__ == "__main__":
    heaters = gf.Component("heaters")
    heater1 = gf.components.straight_heater_metal(length=50)
    heater2 = gf.components.straight_heater_metal(length=50).move([0, 1])

    heaters = gf.Component()
    heaters << heater1
    heaters << heater2
    heaters.show()

    geometry = mesh2D(
        heaters,
        [(25, -2), (25, 25)],
        exclude_layers=[(1, 10)],
        refine_resolution={(1, 0): 0.02, (47, 0): 0.07},
    )

    import gmsh

    gmsh.write("mesh.msh")
    gmsh.clear()
    geometry.__exit__()

    import meshio

    mesh_from_file = meshio.read("mesh.msh")

    def create_mesh(mesh, cell_type, prune_z=False):
        cells = mesh.get_cells_type(cell_type)
        cell_data = mesh.get_cell_data("gmsh:physical", cell_type)
        points = mesh.points[:, :2] if prune_z else mesh.points
        return meshio.Mesh(
            points=points,
            cells={cell_type: cells},
            cell_data={"name_to_read": [cell_data]},
        )

    line_mesh = create_mesh(mesh_from_file, "line", prune_z=True)
    meshio.write("facet_mesh.xdmf", line_mesh)

    triangle_mesh = create_mesh(mesh_from_file, "triangle", prune_z=True)
    meshio.write("mesh.xdmf", triangle_mesh)



"""Returns a switch_tree.

          __
        _|  |_
  __   | |  |_   _
 |  |__| |__|    |
_|  |__          |dy
 |__|  |  __     |
       |_|  |_   |
         |  |_   -
         |__|

   |<-dx->|

"""

import gdsfactory as gf
from gdsfactory.components.mmi2x2 import mmi2x2
from gdsfactory.components.mzi import mzi1x2_2x2
from gdsfactory.components.splitter_tree import splitter_tree
from gdsfactory.components.straight_heater_metal import straight_heater_metal

mzi = gf.partial(
    mzi1x2_2x2,
    combiner=mmi2x2,
    delta_length=0,
    straight_x_top=straight_heater_metal,
    length_x=None,
)

switch_tree = gf.partial(splitter_tree, coupler=mzi, spacing=(500, 100))


if __name__ == "__main__":
    # c = mzi()
    c = switch_tree(noutputs=16)
    c.show(show_ports=True)



"""LVS demo."""

import gdsfactory as gf


@gf.cell
def pads_with_routes(radius: float = 10):
    """Returns MZI interferometer with bend."""
    c = gf.Component()
    pad = gf.components.pad()

    tl = c << pad
    bl = c << pad

    tr = c << pad
    br = c << pad

    tl.move((0, 300))
    br.move((500, 0))
    tr.move((500, 500))

    ports1 = [bl.ports["e3"], tl.ports["e3"]]
    ports2 = [br.ports["e1"], tr.ports["e1"]]
    routes = gf.routing.get_bundle(ports1, ports2, cross_section="metal3")

    for route in routes:
        c.add(route.references)

    return c


if __name__ == "__main__":
    c = pads_with_routes(radius=100)
    c.show(show_ports=True)



"""Classes and utilities for working with KLayout technology files (.lyp, .lyt).

This module will enable conversion between gdsfactory settings and KLayout technology.
"""

import os
import pathlib
import re
from typing import Dict, List, Optional, Set, Tuple, Union

from lxml import etree
from pydantic import BaseModel, Field, validator

from gdsfactory.config import PATH
from gdsfactory.tech import LayerStack

Layer = Tuple[int, int]
ConductorViaConductorName = Tuple[str, str, str]


def append_file_extension(filename: Union[str, pathlib.Path], extension: str) -> str:
    """Try appending extension to file."""
    # Handle whether given with '.'
    if "." not in extension:
        extension = f".{extension}"

    if isinstance(filename, str) and not filename.endswith(extension):
        filename += extension

    if isinstance(filename, pathlib.Path) and not str(filename).endswith(extension):
        filename = filename.with_suffix(extension)
    return filename


class LayerView(BaseModel):
    """KLayout layer properties.

    Docstrings copied from KLayout documentation (with some modifications):
    https://www.klayout.de/lyp_format.html

    Attributes:
        layer: GDSII layer.
        layer_in_name: Whether to display the name as 'name layer/datatype' rather than just the layer.
        width: This is the line width of the frame in pixels (or empty for the default which is 1).
        line_style: This is the number of the line style used to draw the shape boundaries.
            An empty string is "solid line". The values are "Ix" for one of the built-in styles
            where "I0" is "solid", "I1" is "dotted" etc.
        dither_pattern: This is the number of the dither pattern used to fill the shapes.
            The values are "Ix" for one of the built-in pattern where "I0" is "solid" and "I1" is "clear".
        frame_color: The color of the frames.
        fill_color: The color of the fill pattern inside the shapes.
        animation: This is a value indicating the animation mode.
            0 is "none", 1 is "scrolling", 2 is "blinking" and 3 is "inverse blinking".
        fill_brightness: This value modifies the brightness of the fill color. See "frame-brightness".
        frame_brightness: This value modifies the brightness of the frame color.
            0 is unmodified, -100 roughly adds 50% black to the color which +100 roughly adds 50% white.
        xfill: Whether boxes are drawn with a diagonal cross.
        marked: Whether the entry is marked (drawn with small crosses).
        transparent: Whether the entry is transparent.
        visible: Whether the entry is visible.
        valid: Whether the entry is valid. Invalid layers are drawn but you can't select shapes on those layers.
        group_members: Add a list of group members to the LayerView.
    """

    layer: Optional[Layer] = None
    layer_in_name: bool = False
    frame_color: Optional[str] = None
    fill_color: Optional[str] = None
    frame_brightness: Optional[int] = 0
    fill_brightness: Optional[int] = 0
    dither_pattern: Optional[str] = None
    line_style: Optional[str] = None
    valid: bool = True
    visible: bool = True
    transparent: bool = False
    width: Optional[int] = None
    marked: bool = False
    xfill: bool = False
    animation: int = 0
    group_members: Dict[str, "LayerView"] = Field(default_factory=dict)

    def __init__(self, **data):
        """Initialize LayerView object."""
        super().__init__(**data)

        # Iterate through all items, adding group members as needed
        for name, field in self.__fields__.items():
            default = field.get_default()
            if isinstance(default, LayerView):
                self.group_members[name] = default

    @validator("frame_color", "fill_color")
    def color_is_valid(cls, color):
        import numpy as np
        from matplotlib.colors import CSS4_COLORS

        try:
            if color is None:  # not specified
                color = None
            elif np.size(color) == 3:  # in format (0.5, 0.5, 0.5)
                color = np.array(color)
                if np.any(color > 1) or np.any(color < 0):
                    raise ValueError
                color = np.array(np.round(color * 255), dtype=int)
                color = "#{:02x}{:02x}{:02x}".format(*color)
            elif color[0] == "#":  # in format #1d2e3f
                if len(color) != 7:
                    raise ValueError
                int(color[1:], 16)  # Will throw error if not hex format
            else:  # in named format 'gold'
                color = CSS4_COLORS[color.lower()]
        except Exception as error:
            raise ValueError(
                "LayerView color must be specified as a "
                "0-1 RGB triplet, (e.g. [0.5, 0.1, 0.9]), an HTML hex color string "
                "(e.g. '#a31df4'), or a CSS3 color name (e.g. 'gold' or "
                "see http://www.w3schools.com/colors/colors_names.asp )"
            ) from error
        return color

    def __str__(self):
        """Returns a formatted view of properties and their values."""
        return "LayerView:\n\t" + "\n\t".join(
            [f"{k}: {v}" for k, v in self.dict().items()]
        )

    def __repr__(self):
        """Returns a formatted view of properties and their values."""
        return self.__str__()

    def _get_xml_element(self, tag: str, name: str) -> etree.Element:
        """Get XML Element from attributes."""
        prop_keys = [
            "frame-color",
            "fill-color",
            "frame-brightness",
            "fill-brightness",
            "dither-pattern",
            "line-style",
            "valid",
            "visible",
            "transparent",
            "width",
            "marked",
            "xfill",
            "animation",
            "name",
            "source",
        ]
        el = etree.Element(tag)
        for prop_name in prop_keys:
            if prop_name == "source":
                layer = self.layer
                prop_val = f"{layer[0]}/{layer[1]}@1" if layer else "*/*@*"
            elif prop_name == "name":
                prop_val = name
                if self.layer_in_name:
                    prop_val += f"{self.layer[0]}/{self.layer[1]}"
            else:
                prop_val = getattr(self, "_".join(prop_name.split("-")), None)
                if isinstance(prop_val, bool):
                    prop_val = f"{prop_val}".lower()
            subel = etree.SubElement(el, prop_name)
            if prop_val is not None:
                subel.text = str(prop_val)
        return el

    def to_xml(self, name: str) -> etree.Element:
        """Return an XML representation of the LayerView."""
        props = self._get_xml_element("properties", name=name)
        for member_name, member in self.group_members.items():
            props.append(member._get_xml_element("group-members", name=member_name))
        return props


class CustomDitherPattern(BaseModel):
    """Custom dither pattern. See KLayout documentation for more info.

    Attributes:
        order: Order of pattern.
        pattern: Pattern to use.
    """

    order: int
    pattern: str

    def to_xml(self, name: str) -> etree.Element:
        el = etree.Element("custom-dither-pattern")

        subel = etree.SubElement(el, "pattern")
        lines = self.pattern.split("\n")
        if len(lines) == 1:
            subel.text = lines[0]
        else:
            for line in lines:
                etree.SubElement(subel, "line").text = line

        etree.SubElement(el, "order").text = str(self.order)
        etree.SubElement(el, "name").text = name
        return el


class CustomLineStyle(BaseModel):
    """Custom line style. See KLayout documentation for more info.

    Attributes:
        order: Order of pattern.
        pattern: Pattern to use.
    """

    order: int
    pattern: str

    def to_xml(self, name: str) -> etree.Element:
        el = etree.Element("custom-line-pattern")

        etree.SubElement(el, "pattern").text = self.pattern
        etree.SubElement(el, "order").text = str(self.order)
        etree.SubElement(el, "name").text = name
        return el


class CustomPatterns(BaseModel):
    dither_patterns: Dict[str, CustomDitherPattern] = Field(default_factory=dict)
    line_styles: Dict[str, CustomLineStyle] = Field(default_factory=dict)

    def to_yaml(self, filename: str) -> None:
        """Export custom patterns to a yaml file.

        Args:
            filename: Name of output file.
        """
        import yaml

        def _str_presenter(dumper: yaml.Dumper, data: str) -> yaml.ScalarNode:
            if "\n" in data:  # check for multiline string
                return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
            return dumper.represent_scalar("tag:yaml.org,2002:str", data)

        append_file_extension(filename, ".yml")

        yaml.add_representer(str, _str_presenter)
        yaml.representer.SafeRepresenter.add_representer(str, _str_presenter)

        with open(filename, "w") as file:
            file.write(yaml.safe_dump_all([self.dict()], indent=2, sort_keys=False))

    @staticmethod
    def from_yaml(filename: str) -> "CustomPatterns":
        """Import a CustomPatterns object from a yaml file.

        Args:
            filename: Name of output file.
        """
        from omegaconf import OmegaConf

        with open(filename) as file:
            loaded_yml = OmegaConf.to_container(OmegaConf.load(file))
        if not (
            "dither_patterns" in loaded_yml.keys() or "line_styles" in loaded_yml.keys()
        ):
            raise KeyError("The yaml file is not properly formatted.")
        return CustomPatterns(**loaded_yml)

    def __str__(self):
        """Prints the number of each type of custom pattern."""
        return f"CustomPatterns: {len(self.dither_patterns)} dither patterns, {len(self.line_styles)} line styles"


def _process_name(
    name: str, layer_pattern: Union[str, re.Pattern]
) -> Optional[Tuple[str, bool]]:
    """Strip layer info from name if it exists.

    Args:
        name: XML-formatted name entry.
        layer_pattern: Regex pattern to match layers with.
    """
    if not name:
        return None
    layer_in_name = False
    match = re.search(layer_pattern, name)
    if match:
        name = name[: match.start()].strip()
        layer_in_name = True
    return name, layer_in_name


def _process_layer(
    layer: str, layer_pattern: Union[str, re.Pattern]
) -> Optional[Layer]:
    """Convert .lyp XML layer entry to a Layer.

    Args:
        layer: XML-formatted layer entry.
        layer_pattern: Regex pattern to match layers with.
    """
    match = re.search(layer_pattern, layer)
    if not match:
        raise OSError(f"Could not read layer {layer}!")
    v = match.group().split("/")
    return None if v == ["*", "*"] else (int(v[0]), int(v[1]))


def _properties_to_layerview(
    element, layer_pattern: Union[str, re.Pattern]
) -> Optional[Tuple[str, LayerView]]:
    """Read properties from .lyp XML and generate LayerViews from them.

    Args:
        element: XML Element to iterate over.
        layer_pattern: Regex pattern to match layers with.
    """
    prop_dict = {"layer_in_name": False}
    name = ""
    for prop in element.iterchildren():
        prop_tag = prop.tag
        if prop_tag == "name":
            name = _process_name(prop.text, layer_pattern)
            if name is None:
                return None
            name, layer_in_name = name
            prop_dict["layer_in_name"] = layer_in_name
            continue
        elif prop_tag == "source":
            val = _process_layer(prop.text, layer_pattern)
            prop_tag = "layer"
        elif prop_tag == "group-members":
            val = [
                _properties_to_layerview(e, layer_pattern)
                for e in element.iterchildren("group-members")
            ]
        else:
            val = prop.text
        prop_tag = "_".join(prop_tag.split("-"))
        prop_dict[prop_tag] = val
    return name, LayerView(**prop_dict)


class LayerDisplayProperties(BaseModel):
    """A container for layer properties for KLayout layer property (.lyp) files.

    Attributes:
        layer_views: Dictionary of LayerViews describing how to display gds layers.
        custom_patterns: CustomPatterns object containing custom dither patterns and line styles.
    """

    layer_views: Dict[str, LayerView] = Field(default_factory=dict)
    custom_patterns: Optional[CustomPatterns] = None

    def __init__(self, **data):
        """Initialize LayerDisplayProperties object."""
        super().__init__(**data)

        for field in self.dict():
            val = getattr(self, field)
            if isinstance(val, LayerView):
                if val.name is None:
                    val.name = field
                self.add_layer_view(val)

    def add_layer_view(self, name: str, layer_view: Optional[LayerView]) -> None:
        """Adds a layer to LayerDisplayProperties.

        Args:
            name: Name of the LayerView.
            layer_view: LayerView to add.
        """
        if name in self.layer_views:
            raise ValueError(
                f"Adding {name!r} already defined {list(self.layer_views.keys())}"
            )
        else:
            self.layer_views[name] = layer_view

    def get_layer_views(self, exclude_groups: bool = False) -> Dict[str, LayerView]:
        """Return all LayerViews.

        Args:
            exclude_groups: Whether to exclude LayerViews that contain other LayerViews.
        """
        layers = {}
        for name, view in self.layer_views.items():
            if view.group_members and not exclude_groups:
                for member_name, member in view.group_members.items():
                    layers[member_name] = member
                continue
            layers[name] = view
        return layers

    def get_layer_view_groups(self) -> Dict[str, LayerView]:
        """Return the LayerViews that contain other LayerViews."""
        return {name: lv for name, lv in self.layer_views.items() if lv.group_members}

    def __str__(self) -> str:
        """Prints the number of LayerView objects in the LayerDisplayProperties object."""
        lvs = self.get_layer_views()
        groups = self.get_layer_view_groups()
        return (
            f"LayerDisplayProperties: {len(lvs)} layers ({len(groups)} groups)\n"
            f"\t{self.custom_patterns}"
        )

    def get(self, name: str) -> LayerView:
        """Returns Layer from name.

        Args:
            name: Name of layer.
        """
        if name not in self.layer_views:
            raise ValueError(f"Layer {name!r} not in {list(self.layer_views.keys())}")
        else:
            return self.layer_views[name]

    def __getitem__(self, val: str):
        """Allows accessing to the layer names like ls['gold2'].

        Args:
            val: Layer name to access within the LayerDisplayProperties.

        Returns:
            self.layers[val]: LayerView in the LayerDisplayProperties.

        """
        try:
            return self.layer_views[val]
        except Exception as error:
            raise ValueError(
                f"LayerView {val!r} not in LayerDisplayProperties {list(self.layer_views.keys())}"
            ) from error

    def get_from_tuple(self, layer_tuple: Layer) -> LayerView:
        """Returns LayerView from layer tuple.

        Args:
            layer_tuple: Tuple of (gds_layer, gds_datatype).

        Returns:
            LayerView
        """
        tuple_to_name = {v.layer: k for k, v in self.layer_views.items()}
        if layer_tuple not in tuple_to_name:
            raise ValueError(
                f"LayerView {layer_tuple} not in {list(tuple_to_name.keys())}"
            )

        name = tuple_to_name[layer_tuple]
        return self.layer_views[name]

    def get_layer_tuples(self) -> Set[Layer]:
        """Returns a tuple for each layer."""
        return {layer.layer for layer in self.get_layer_views().values()}

    def to_lyp(self, filepath: str, overwrite: bool = True) -> None:
        """Write all layer properties to a KLayout .lyp file.

        Args:
            filepath: to write the .lyp file to (appends .lyp extension if not present).
            overwrite: Whether to overwrite an existing file located at the filepath.
        """
        filepath = pathlib.Path(filepath)
        filepath = append_file_extension(filepath, ".lyp")

        if os.path.exists(filepath) and not overwrite:
            raise OSError("File exists, cannot write.")

        root = etree.Element("layer-properties")

        for name, lv in self.layer_views.items():
            root.append(lv.to_xml(name))

        for name, dp in self.custom_patterns.dither_patterns.items():
            root.append(dp.to_xml(name))

        for name, ls in self.custom_patterns.line_styles.items():
            root.append(ls.to_xml(name))

        with open(filepath, "wb") as file:
            file.write(
                etree.tostring(
                    root, encoding="utf-8", pretty_print=True, xml_declaration=True
                )
            )

    @staticmethod
    def from_lyp(
        filepath: str, layer_pattern: Optional[Union[str, re.Pattern]] = None
    ) -> "LayerDisplayProperties":
        r"""Write all layer properties to a KLayout .lyp file.

        Args:
            filepath: to write the .lyp file to (appends .lyp extension if not present).
            layer_pattern: Regex pattern to match layers with. Defaults to r'(\d+|\*)/(\d+|\*)'.
        """
        layer_pattern = re.compile(layer_pattern or r"(\d+|\*)/(\d+|\*)")

        filepath = append_file_extension(filepath, ".lyp")

        if not os.path.exists(filepath):
            raise OSError("File not found!")

        tree = etree.parse(filepath)
        root = tree.getroot()
        if root.tag != "layer-properties":
            raise OSError("Layer properties file incorrectly formatted, cannot read.")

        layer_views = {}
        for layer_block in root.iter("properties"):
            name, lv = _properties_to_layerview(
                layer_block, layer_pattern=layer_pattern
            )
            if lv:
                layer_views[name] = lv

        custom_dither_patterns = {}
        for dither_block in root.iterchildren("custom-dither-pattern"):
            name = dither_block.find("name").text
            if name is None:
                continue

            custom_dither_patterns[name] = CustomDitherPattern(
                order=dither_block.find("order").text,
                pattern="\n".join(
                    [line.text for line in dither_block.find("pattern").iterchildren()]
                ),
            )
        custom_line_styles = {}
        for line_block in root.iterchildren("custom-line-style"):
            name = line_block.find("name").text
            if name is None:
                continue
            custom_line_styles[name] = CustomLineStyle(
                order=line_block.find("order").text,
                pattern=line_block.find("pattern").text,
            )
        custom_patterns = CustomPatterns(
            dither_patterns=custom_dither_patterns, line_styles=custom_line_styles
        )

        return LayerDisplayProperties(
            layer_views=layer_views, custom_patterns=custom_patterns
        )

    def to_yaml(self, layer_file: str, pattern_file: Optional[str] = None) -> None:
        """Export layer properties to two yaml files.

        Args:
            layer_file: Name of the file to write LayerViews to.
            pattern_file: Name of the file to write custom dither patterns and line styles to.
        """
        import yaml

        append_file_extension(layer_file, ".yml")
        lvs = {name: lv.dict() for name, lv in self.layer_views.items()}
        with open(layer_file, "w") as lf:
            lf.write(yaml.safe_dump_all([lvs], indent=2, sort_keys=False))

        if pattern_file:
            append_file_extension(pattern_file, ".yml")
            self.custom_patterns.to_yaml(pattern_file)

    @staticmethod
    def from_yaml(
        layer_file: str = None, pattern_file: Optional[str] = None
    ) -> "LayerDisplayProperties":
        """Import layer properties from two yaml files.

        Args:
            layer_file: Name of the file to read LayerViews from.
            pattern_file: Name of the file to read custom dither patterns and line styles from.
        """
        from omegaconf import OmegaConf

        append_file_extension(layer_file, ".yml")

        with open(layer_file) as lf:
            layers = OmegaConf.to_container(OmegaConf.load(lf))
        props = LayerDisplayProperties(
            layer_views={name: LayerView(**lv) for name, lv in layers.items()}
        )
        if pattern_file:
            append_file_extension(pattern_file, ".yml")
            props.custom_patterns = CustomPatterns.from_yaml(pattern_file)
        return props


class KLayoutTechnology(BaseModel):
    """A container for working with KLayout technologies (requires klayout Python package).

    Useful for importing/exporting Layer Properties (.lyp) and Technology (.lyt) files.

    Properties:
        layer_properties: Defines all the layer display properties needed for a .lyp file from LayerView objects.
        technology: KLayout Technology object from the KLayout API. Set name, dbu, etc.
        connectivity: List of layer names connectivity for netlist tracing.
    """

    # TODO: Add import method
    import klayout.db as db

    name: str
    layer_properties: Optional[LayerDisplayProperties] = None
    technology: db.Technology = Field(default_factory=db.Technology)
    connectivity: Optional[List[ConductorViaConductorName]] = None

    def export_technology_files(
        self,
        tech_dir: str,
        lyp_filename: str = "layers",
        lyt_filename: str = "tech",
        layer_stack: Optional[LayerStack] = None,
        mebes_config: Optional[dict] = None,
    ) -> None:
        """Write technology files into 'tech_dir'.

        Args:
            tech_dir: Where to write the technology files to.
            lyp_filename: Name of the layer properties file.
            lyt_filename: Name of the layer technology file.
            layer_stack: If specified, write a 2.5D section in the technology file based on the LayerStack.
            mebes_config: A dictionary specifying the KLayout mebes reader config.
        """
        # Format file names if necessary
        lyp_filename = append_file_extension(lyp_filename, ".lyp")
        lyt_filename = append_file_extension(lyt_filename, ".lyt")

        tech_path = pathlib.Path(tech_dir)
        lyp_path = tech_path / lyp_filename
        lyt_path = tech_path / lyt_filename

        # Specify relative file name for layer properties file
        self.technology.layer_properties_file = lyp_filename

        if not self.technology.name:
            self.technology.name = self.name

        # TODO: Also interop with xs scripts?

        # Write lyp to file
        self.layer_properties.to_lyp(lyp_path)

        root = etree.XML(self.technology.to_xml().encode("utf-8"))

        # KLayout tech doesn't include mebes config, so add it after lefdef config:
        if not mebes_config:
            mebes_config = {
                "invert": False,
                "subresolution": True,
                "produce-boundary": True,
                "num-stripes-per-cell": 64,
                "num-shapes-per-cell": 0,
                "data-layer": 1,
                "data-datatype": 0,
                "data-name": "DATA",
                "boundary-layer": 0,
                "boundary-datatype": 0,
                "boundary-name": "BORDER",
                "layer-map": "layer_map()",
                "create-other-layers": True,
            }
        mebes = etree.Element("mebes")
        for k, v in mebes_config.items():
            v = str(v).lower() if isinstance(v, bool) else str(v)
            etree.SubElement(mebes, k).text = v

        reader_opts = root.find("reader-options")
        lefdef_idx = reader_opts.index(reader_opts.find("lefdef"))
        reader_opts.insert(lefdef_idx + 1, mebes)

        if layer_stack is not None:
            # KLayout 0.27.x won't have a way to read/write the 2.5D info for technologies, so add manually
            # Should be easier in 0.28.x
            d25_element = [e for e in list(root) if e.tag == "d25"]
            if len(d25_element) != 1:
                raise KeyError("Could not get a single index for the d25 element.")
            d25_element = d25_element[0]

            src_element = [e for e in list(d25_element) if e.tag == "src"]
            if len(src_element) != 1:
                raise KeyError("Could not get a single index for the src element.")
            src_element = src_element[0]

            for layer_level in layer_stack.layers.values():
                # Round the float based on the database unit (dbu) to not end up with numbers like: 2.0700000000000003
                rounding_place = len(str(self.technology.dbu).split(".")[-1])
                zmin = round(layer_level.zmin, rounding_place)
                zmax = round(zmin + layer_level.thickness, rounding_place)
                src_element.text += (
                    f"{layer_level.layer[0]}/{layer_level.layer[1]}: {zmin} {zmax}\n"
                )

        # root['connectivity']['connection']= '41/0,44/0,45/0'
        if self.connectivity is not None:
            src_element = [e for e in list(root) if e.tag == "connectivity"]
            if len(src_element) != 1:
                raise KeyError("Could not get a single index for the src element.")
            src_element = src_element[0]
            for layer_name_c1, layer_name_via, layer_name_c2 in self.connectivity:
                layer_c1 = self.layer_properties.layer_views[layer_name_c1].layer
                layer_via = self.layer_properties.layer_views[layer_name_via].layer
                layer_c2 = self.layer_properties.layer_views[layer_name_c2].layer
                connection = ",".join(
                    [
                        f"{layer[0]}/{layer[1]}"
                        for layer in [layer_c1, layer_via, layer_c2]
                    ]
                )

                etree.SubElement(src_element, "connection").text = connection

        # The indentation can easily get messed up when adding elements to existing XML objects, so we need to filter it
        # by passing it through another parser that strips it of any whitespace and then using pretty_print to re-format
        # the indentation.
        parser = etree.XMLParser(remove_blank_text=True)

        script = etree.tostring(
            etree.XML(etree.tostring(root), parser),
            encoding="utf-8",
            pretty_print=True,
            xml_declaration=True,
        )

        # Write lyt to file
        lyt_path.write_bytes(script)

    class Config:
        """Allow db.Technology type."""

        arbitrary_types_allowed = True


LAYER_PROPERTIES = LayerDisplayProperties.from_lyp(str(PATH.klayout_lyp))


def yaml_test():
    tech_dir = PATH.repo / "extra" / "test_tech"

    # Load from existing layer properties file
    lyp = LayerDisplayProperties.from_lyp(str(PATH.klayout_lyp))
    print("Loaded from .lyp", lyp)

    # Export layer properties to yaml files
    layer_yaml = str(tech_dir / "layers.yml")
    pattern_yaml = str(tech_dir / "patterns.yml")
    lyp.to_yaml(layer_yaml, pattern_yaml)

    # Load layer properties from yaml files and check that they're the same
    lyp_loaded = LayerDisplayProperties.from_yaml(layer_yaml, pattern_yaml)
    print("Loaded from .yaml", lyp_loaded)

    assert lyp_loaded == lyp


if __name__ == "__main__":
    from gdsfactory.tech import LAYER_STACK

    lyp = LayerDisplayProperties.from_lyp(str(PATH.klayout_lyp))

    # str_xml = open(PATH.klayout_tech / "tech.lyt").read()
    # new_tech = db.Technology.technology_from_xml(str_xml)
    # generic_tech = KLayoutTechnology(layer_properties=lyp)
    #
    connectivity = [("M1", "VIA1", "M2"), ("M2", "VIA2", "M3")]

    c = generic_tech = KLayoutTechnology(
        name="generic", layer_properties=lyp, connectivity=connectivity
    )
    tech_dir = PATH.repo / "extra" / "test_tech"
    # tech_dir = pathlib.Path("/home/jmatres/.klayout/salt/gdsfactory/tech/")
    tech_dir.mkdir(exist_ok=True, parents=True)

    generic_tech.export_technology_files(tech_dir=tech_dir, layer_stack=LAYER_STACK)

    yaml_test()



from typing import Dict, Optional

import numpy as np
import pandas as pd
from omegaconf import DictConfig, OmegaConf

import gdsfactory as gf
from gdsfactory.component_layout import _rotate_points
from gdsfactory.name import clean_name
from gdsfactory.types import PathType


def read_labels_yaml(
    csvpath: PathType, prefix: Optional[str] = None
) -> Dict[str, DictConfig]:
    """Read labels from csvfile in YAML format."""
    labels = pd.read_csv(csvpath)
    cells = OmegaConf.create()

    for label in labels.iterrows():
        label = label[1]
        if prefix and not label[0].startswith(prefix):
            continue

        d = OmegaConf.create(label[0])
        x = label[1]
        y = label[2]
        rotation = label[3]
        cell_name = d["component_name"]
        instance_name = clean_name(f"{cell_name}_{x}_{y}")
        # wavelength = d.get("wavelength", 1.55)
        # polarization = d["polarization"]
        # ports = d["ports"]

        d["x"] = x
        d["y"] = y
        d["rotation"] = rotation
        cells[instance_name] = d
    return cells


def add_port_markers(gdspath, csvpath, marker_size=20, marker_layer=(203, 0)):
    """Add port markers from port info extracted from a gdspath and csvpath."""
    c = gf.Component("overlay")
    c << gf.import_gds(gdspath)
    cells = read_labels_yaml(csvpath=csvpath)

    for cell in cells.values():
        for port in cell["ports"].values():
            r = c << gf.components.rectangle(
                size=(marker_size, marker_size), layer=marker_layer
            )

            x = port["center"][0]
            y = port["center"][1]
            rotation = cell["rotation"]
            x, y = np.array(_rotate_points((x, y), angle=rotation)).flatten()

            r.x = x + cell["x"]
            r.y = y + cell["y"]
    return c


if __name__ == "__main__":
    # c = read_labels_yaml(csvpath="/home/jmatres/mask.csv", prefix="component_name")
    # csvpath = "/home/jmatres/mask.csv"
    # prefix = "component_name"

    # labels = pd.read_csv(csvpath)
    # cells = OmegaConf.create()

    # for label in labels.iterrows():
    #     label = label[1]
    #     if prefix and not label[0].startswith(prefix):
    #         continue

    #     d = OmegaConf.create(label[0])
    #     x = label[1]
    #     y = label[2]
    #     rotation = label[3]
    #     cell_name = d["component_name"]
    #     wavelength = d.get("wavelength", 1.55)
    #     polarization = d["polarization"]
    #     ports = d["ports"]
    #     instance_name = clean_name(f"{cell_name}_{x}_{y}")
    #     cells[instance_name] = d

    csvpath = "/home/jmatres/mask.csv"
    gdspath = "/home/jmatres/mask.gds"
    marker_size = 40
    marker_layer = (203, 0)

    c = gf.Component("overlay")
    c << gf.import_gds(gdspath)
    cells = read_labels_yaml(csvpath=csvpath)

    for cell in cells.values():
        for port in cell["ports"].values():
            r = c << gf.components.rectangle(
                size=(marker_size, marker_size), layer=marker_layer
            )

            x = port["center"][0]
            y = port["center"][1]
            rotation = cell["rotation"]
            x, y = np.array(_rotate_points((x, y), angle=rotation)).flatten()

            r.x = x + cell["x"]
            r.y = y + cell["y"]

    c.show(show_ports=True)
